# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-05 08:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0002_auto_20190527_0938'),
    ]

    operations = [
        migrations.AlterField(
            model_name='message',
            name='component',
            field=models.CharField(choices=[(b'processing', 'Processing action'), (b'planogram_import', 'Planogram import'), (b'internal', 'Internal')], max_length=64),
        ),
        migrations.AlterField(
            model_name='message',
            name='delivery_status',
            field=models.CharField(choices=[(b'message_pending', 'Pending'), (b'message_delivered', 'Delivered'), (b'message_rejected', 'Rejected'), (b'message_failed', 'Failed')], max_length=64),
        ),
    ]
