# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-07 12:55
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0003_auto_20190605_0816'),
        ('notifications', '0003_auto_20190605_0817'),
    ]

    operations = [
    ]
