# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-27 06:02
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='WebhookRouter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.URLField()),
                ('auth_token', models.CharField(editable=False, max_length=64)),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='Date Modified')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='webhooks', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Webhook router',
                'verbose_name_plural': 'Webhook routers',
            },
        ),
    ]
